# pylint: disable=R0902
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


class CacheWrapper:
    def __init__(
        self, file_path, cache_type, database_adaptor, tentacles_requirements, **kwargs
    ):
        self.file_path = file_path
        self.cache_type = cache_type
        self.database_adaptor = database_adaptor
        self.db_kwargs = kwargs
        self._cache_database = None
        self._db_path = None
        self.previous_db_metadata = None
        self.tentacles_requirements = tentacles_requirements.summary()

    def get_database(self) -> tuple:
        """
        Returns the database, creates it if messing
        """
        if self._cache_database is None:
            self._cache_database = self.cache_type(
                self.file_path, database_adaptor=self.database_adaptor, **self.db_kwargs
            )
            self._db_path = self._cache_database.get_db_path()
            return self._cache_database, True
        return self._cache_database, False

    def is_open(self):
        """
        :return: True if a database is open
        """
        return self._cache_database is not None

    async def close(self):
        """
        Closes the current database. Stores its metadata into self.previous_db_metadata
        """
        if self.is_open():
            self.previous_db_metadata = self._cache_database.get_non_default_metadata()
            await self._cache_database.close()
            self._cache_database = None
            return True
        return False

    async def clear(self):
        """
        Clears the database, deleting its data
        """
        if self._cache_database is not None:
            await self._cache_database.clear()

    def get_path(self):
        """
        :return: the database path
        """
        return self._db_path



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import copy
import json
import shutil
import pytest
import mock
import octobot_commons.json_util
import octobot_commons.profiles as profiles
import octobot_commons.constants as constants
import octobot_commons.enums as enums
import octobot_commons.tests.test_config as test_config

from tests.profiles import profile, get_profile_path, get_profiles_path


def test_read_config(profile):
    save_ref = profile
    assert profile.read_config() is save_ref
    assert profile.profile_id == "default"
    assert profile.name == "default"
    assert profile.description == "OctoBot default profile."
    assert profile.avatar == "default_profile.png"
    assert profile.avatar_path == os.path.join(test_config.TEST_CONFIG_FOLDER, "default_profile.png")
    assert profile.origin_url == "https://default.url"
    assert len(profile.config) == 5
    assert isinstance(profile.config, dict)

    profile.path = ""
    with pytest.raises(FileNotFoundError):
        profile.read_config()


def test_save_config(profile):
    with mock.patch.object(profile, "validate_and_save_config", mock.Mock()) as validate_and_save_config_mock, \
            mock.patch.object(profile, "_filter_fill_elements", mock.Mock()) as _filter_fill_elements_mock:
        profile.config = {}
        # nothing to operate on
        global_config = {}
        profile.save_config(global_config)
        assert profile.config == {}
        validate_and_save_config_mock.assert_called_once()
        _filter_fill_elements_mock.assert_not_called()

        validate_and_save_config_mock.reset_mock()
        _filter_fill_elements_mock.reset_mock()
        profile.config = {}
        # things in config
        global_config = {
            profile.FULLY_MANAGED_ELEMENTS[0]: "plop",
            profile.FULLY_MANAGED_ELEMENTS[1]: "plip",
            "stuff": "plip"
        }
        profile.save_config(global_config)
        assert profile.config == {
            profile.FULLY_MANAGED_ELEMENTS[0]: "plop",
            profile.FULLY_MANAGED_ELEMENTS[1]: "plip"
        }
        validate_and_save_config_mock.assert_called_once()
        _filter_fill_elements_mock.assert_not_called()

        validate_and_save_config_mock.reset_mock()
        _filter_fill_elements_mock.reset_mock()
        profile.config = {}
        # things in config
        global_config = {
            profile.FULLY_MANAGED_ELEMENTS[0]: "plop",
            profile.FULLY_MANAGED_ELEMENTS[1]: "plip",
            "stuff": "plip",
            next(iter(profile.PARTIALLY_MANAGED_ELEMENTS)): {"ploup": True},
        }
        profile.save_config(global_config)
        assert profile.config == {
            profile.FULLY_MANAGED_ELEMENTS[0]: "plop",
            profile.FULLY_MANAGED_ELEMENTS[1]: "plip",
        }
        validate_and_save_config_mock.assert_called_once()
        _filter_fill_elements_mock.assert_called_once_with(global_config,
                                                           profile.config,
                                                           next(iter(profile.PARTIALLY_MANAGED_ELEMENTS)),
                                                           profile.PARTIALLY_MANAGED_ELEMENTS_ALLOWED_KEYS[
                                                               next(iter(profile.PARTIALLY_MANAGED_ELEMENTS))
                                                           ])


def test_validate(profile):
    with mock.patch.object(octobot_commons.json_util, "validate", mock.Mock()) as validate_mock:
        profile.validate()
        validate_mock.assert_called_once_with(profile.as_dict(), profile.schema_path)


def test_validate_and_save_config(profile):
    save_file = "profile_config.json"
    with mock.patch.object(profile, "validate", mock.Mock()) as validate_mock, \
            mock.patch.object(profile, "config_file", mock.Mock(return_value=save_file)), \
            mock.patch.object(profile, "save", mock.Mock()) as save_mock:
        profile.validate_and_save_config()
        validate_mock.assert_called_once()
        save_mock.assert_called_once()


def test_save(profile):
    save_file = "profile_config.json"
    if os.path.isfile(save_file):
        os.remove(save_file)
    try:
        profile.read_config()
        with mock.patch.object(profile, "config_file", mock.Mock(return_value=save_file)):
            profile.save()
            with open(save_file) as config_file:
                saved_profile = json.load(config_file)
            assert saved_profile == profile.as_dict()
    finally:
        if os.path.isfile(save_file):
            os.remove(save_file)


def test_duplicate(profile):
    with mock.patch.object(shutil, "copytree", mock.Mock()) as copytree_mock, \
            mock.patch.object(profiles.Profile, "save", mock.Mock()) as save_mock:
        profile.read_only = True
        profile.imported = True
        profile.origin_url = "hello"
        clone = profile.duplicate()
        assert clone.name == profile.name
        assert clone.description == profile.description
        assert clone.profile_id != profile.description
        assert clone.path != profile.path
        assert clone.profile_id in clone.path
        assert clone.profile_id is not None
        # duplicates are not read_only
        assert clone.read_only is False
        # duplicates are never imported nor have an origin url
        assert clone.imported is False
        assert clone.origin_url is None
        copytree_mock.assert_called_with(profile.path, clone.path)
        save_mock.assert_called_once()

        clone = profile.duplicate(name="123", description="456")
        assert clone.name == "123"
        assert clone.name != profile.name
        assert clone.description == "456"
        assert clone.description != profile.description


def test_as_dict(profile):
    assert profile.as_dict() == {
        constants.CONFIG_PROFILE: {
            constants.CONFIG_ID: None,
            constants.CONFIG_NAME: None,
            constants.CONFIG_DESCRIPTION: None,
            constants.CONFIG_AVATAR: None,
            constants.CONFIG_ORIGIN_URL: None,
            constants.CONFIG_READ_ONLY: False,
            constants.CONFIG_IMPORTED: False,
            constants.CONFIG_COMPLEXITY: enums.ProfileComplexity.MEDIUM.value,
            constants.CONFIG_RISK: enums.ProfileRisk.MODERATE.value,
            constants.CONFIG_TYPE: enums.ProfileType.LIVE.value,
        },
        constants.PROFILE_CONFIG: {},
    }
    profile.read_config()
    # do not test read config
    profile.config = {"a": 1}
    profile.imported = True
    profile.complexity = enums.ProfileComplexity.DIFFICULT
    profile.risk = enums.ProfileRisk.LOW
    profile.profile_type = enums.ProfileType.BACKTESTING
    assert profile.as_dict() == {
        constants.CONFIG_PROFILE: {
            constants.CONFIG_ID: "default",
            constants.CONFIG_NAME: "default",
            constants.CONFIG_DESCRIPTION: "OctoBot default profile.",
            constants.CONFIG_AVATAR: "default_profile.png",
            constants.CONFIG_ORIGIN_URL: "https://default.url",
            constants.CONFIG_READ_ONLY: False,
            constants.CONFIG_IMPORTED: True,
            constants.CONFIG_COMPLEXITY: enums.ProfileComplexity.DIFFICULT.value,
            constants.CONFIG_RISK: enums.ProfileRisk.LOW.value,
            constants.CONFIG_TYPE: enums.ProfileType.BACKTESTING.value,
        },
        constants.PROFILE_CONFIG: {
            "a": 1
        },
    }


def test_config_file(profile):
    assert profile.config_file() == os.path.join(get_profile_path(), constants.PROFILE_CONFIG_FILE)


def test_merge_partially_managed_element_into_config(profile):
    with mock.patch.object(profiles.Profile, "_merge_partially_managed_element", mock.Mock()) as _merge_mock:
        config = {}
        profile.merge_partially_managed_element_into_config(config, constants.CONFIG_EXCHANGES)
        _merge_mock.assert_called_once_with(config,
                                            profile.config,
                                            constants.CONFIG_EXCHANGES,
                                            profile.PARTIALLY_MANAGED_ELEMENTS[constants.CONFIG_EXCHANGES])


def test_merge_partially_managed_element(profile):
    profile.read_config()
    element = next(iter(profile.PARTIALLY_MANAGED_ELEMENTS))
    template = profile.PARTIALLY_MANAGED_ELEMENTS[element]
    config = {
        constants.CONFIG_EXCHANGES: {
            "binance": {
                constants.CONFIG_EXCHANGE_KEY: constants.DEFAULT_API_KEY,
                constants.CONFIG_EXCHANGE_SECRET: constants.DEFAULT_API_SECRET,
            }
        }
    }
    # add constants.CONFIG_ENABLED_OPTION
    profile._merge_partially_managed_element(config, profile.config, element, template)
    assert config == {
        constants.CONFIG_EXCHANGES: {
            "binance": {
                constants.CONFIG_EXCHANGE_KEY: constants.DEFAULT_API_KEY,
                constants.CONFIG_EXCHANGE_SECRET: constants.DEFAULT_API_SECRET,
                constants.CONFIG_ENABLED_OPTION: True
            }
        }
    }
    config = {
        constants.CONFIG_EXCHANGES: {}
    }
    profile.config[constants.CONFIG_EXCHANGES]["binance"][constants.CONFIG_ENABLED_OPTION] = False
    # add whole exchange
    profile._merge_partially_managed_element(config, profile.config, element, template)
    assert config == {
        constants.CONFIG_EXCHANGES: {
            "binance": {
                constants.CONFIG_EXCHANGE_KEY: constants.DEFAULT_API_KEY,
                constants.CONFIG_EXCHANGE_SECRET: constants.DEFAULT_API_SECRET,
                constants.CONFIG_EXCHANGE_PASSWORD: constants.DEFAULT_API_PASSWORD,
                constants.CONFIG_EXCHANGE_TYPE: constants.DEFAULT_EXCHANGE_TYPE,
                constants.CONFIG_ENABLED_OPTION: False
            }
        }
    }
    config = {}
    # add whole exchange and exchanges key with 2 exchanges in profile
    profile.config[constants.CONFIG_EXCHANGES]["kucoin"] = {
        constants.CONFIG_ENABLED_OPTION: True,
        constants.CONFIG_EXCHANGE_TYPE: constants.CONFIG_EXCHANGE_FUTURE
    }
    profile._merge_partially_managed_element(config, profile.config, element, template)
    assert config == {
        constants.CONFIG_EXCHANGES: {
            "binance": {
                constants.CONFIG_EXCHANGE_KEY: constants.DEFAULT_API_KEY,
                constants.CONFIG_EXCHANGE_SECRET: constants.DEFAULT_API_SECRET,
                constants.CONFIG_EXCHANGE_PASSWORD: constants.DEFAULT_API_PASSWORD,
                constants.CONFIG_EXCHANGE_TYPE: constants.DEFAULT_EXCHANGE_TYPE,
                constants.CONFIG_ENABLED_OPTION: False
            },
            "kucoin": {
                constants.CONFIG_EXCHANGE_KEY: constants.DEFAULT_API_KEY,
                constants.CONFIG_EXCHANGE_SECRET: constants.DEFAULT_API_SECRET,
                constants.CONFIG_EXCHANGE_PASSWORD: constants.DEFAULT_API_PASSWORD,
                constants.CONFIG_EXCHANGE_TYPE: constants.CONFIG_EXCHANGE_FUTURE,
                constants.CONFIG_ENABLED_OPTION: True
            }
        }
    }
    config = {
        constants.CONFIG_EXCHANGES: {
            "binance": {
                constants.CONFIG_EXCHANGE_KEY: constants.DEFAULT_API_KEY,
                constants.CONFIG_EXCHANGE_SECRET: constants.DEFAULT_API_SECRET,
                constants.CONFIG_ENABLED_OPTION: True
            }
        }
    }
    # add constants.CONFIG_ENABLED_OPTION with 2 exchanges in profile, update constants.CONFIG_ENABLED_OPTION
    profile._merge_partially_managed_element(config, profile.config, element, template)
    assert config == {
        constants.CONFIG_EXCHANGES: {
            "binance": {
                constants.CONFIG_EXCHANGE_KEY: constants.DEFAULT_API_KEY,
                constants.CONFIG_EXCHANGE_SECRET: constants.DEFAULT_API_SECRET,
                constants.CONFIG_ENABLED_OPTION: False
            },
            "kucoin": {
                constants.CONFIG_EXCHANGE_KEY: constants.DEFAULT_API_KEY,
                constants.CONFIG_EXCHANGE_SECRET: constants.DEFAULT_API_SECRET,
                constants.CONFIG_EXCHANGE_PASSWORD: constants.DEFAULT_API_PASSWORD,
                constants.CONFIG_EXCHANGE_TYPE: constants.CONFIG_EXCHANGE_FUTURE,
                constants.CONFIG_ENABLED_OPTION: True
            }
        }
    }


def test_remove_deleted_elements(profile):
    profile.read_config()
    element = next(iter(profile.PARTIALLY_MANAGED_ELEMENTS))
    config = {
        constants.CONFIG_EXCHANGES: {
            "binance": {
                constants.CONFIG_EXCHANGE_KEY: constants.DEFAULT_API_KEY,
                constants.CONFIG_EXCHANGE_SECRET: constants.DEFAULT_API_SECRET,
                constants.CONFIG_ENABLED_OPTION: True,
            }
        }
    }
    before_sync_elements_count = len(profile.config[element])
    profile.remove_deleted_elements(config)
    # did not remove any element
    assert before_sync_elements_count == len(profile.config[element])
    profile.config[element]["plop"] = config[constants.CONFIG_EXCHANGES]["binance"]
    assert len(profile.config[element]) == before_sync_elements_count + 1
    profile.remove_deleted_elements(config)
    assert before_sync_elements_count == len(profile.config[element])
    # removed "plop" element
    assert list(profile.config[element]) == ["binance"]


def test_get_element_from_template(profile):
    element = next(iter(profile.PARTIALLY_MANAGED_ELEMENTS))
    template = profile.PARTIALLY_MANAGED_ELEMENTS[element]
    template_copy = copy.deepcopy(template)

    template_copy["plop"] = 1
    assert profile._get_element_from_template(template, {"plop": 1}) == template_copy
    assert "plop" not in template


def test_filter_fill_elements(profile):
    profile.read_config()
    config = {
        constants.CONFIG_EXCHANGES: {
            "binance": {
                constants.CONFIG_EXCHANGE_KEY: constants.DEFAULT_API_KEY,
                constants.CONFIG_EXCHANGE_SECRET: constants.DEFAULT_API_SECRET,
                constants.CONFIG_EXCHANGE_PASSWORD: constants.DEFAULT_API_PASSWORD,
                constants.CONFIG_ENABLED_OPTION: True
            }
        }
    }
    allowed_keys = [constants.CONFIG_ENABLED_OPTION, constants.CONFIG_EXCHANGE_PASSWORD]
    profile._filter_fill_elements(config, profile.config, constants.CONFIG_EXCHANGES, allowed_keys)
    assert profile.config[constants.CONFIG_EXCHANGES] == {
        "binance": {
            constants.CONFIG_EXCHANGE_PASSWORD: constants.DEFAULT_API_PASSWORD,
            constants.CONFIG_ENABLED_OPTION: True
        }
    }


def test_get_all_profiles():
    with mock.patch.object(profiles.Profile, "_load_profile", mock.Mock()) as _load_profile_mock:
        nb_files = len(os.listdir(get_profiles_path()))
        assert nb_files > 1
        profiles.Profile.get_all_profiles(get_profiles_path())
        assert _load_profile_mock.call_count == nb_files


def test_load_profile():
    schema_path = "schema_path"
    with mock.patch.object(profiles.Profile, "read_config", mock.Mock()) as read_config_mock:
        profile = profiles.Profile._load_profile(test_config.TEST_CONFIG_FOLDER, schema_path)
        assert profile.path == test_config.TEST_CONFIG_FOLDER
        assert profile.schema_path == schema_path
        read_config_mock.assert_called_once()


def test_get_existing_profiles_ids(profile):
    assert profiles.Profile.get_all_profiles_ids(get_profiles_path()) == ["default"]
    assert profiles.Profile.get_all_profiles_ids(get_profiles_path(), ignore=profile.path) == []



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import re
import functools

import octobot_commons


# pylint: disable=R0902
class Symbol:
    FULL_SYMBOL_GROUPS_REGEX = r"([^//]*)\/([^:]*):?([^-]*)-?([^-]*)-?([^-]*)-?([^-]*)"
    #                             base   /  quote : settlement-identifier-strike price-type
    # Inspired from CCXT https://docs.ccxt.com/en/latest/manual.html#option:
    # //
    # // base asset or currency
    # // ↓
    # // ↓  quote asset or currency
    # // ↓  ↓
    # // ↓  ↓    settlement asset or currency
    # // ↓  ↓    ↓
    # // ↓  ↓    ↓       identifier (settlement date)
    # // ↓  ↓    ↓       ↓
    # // ↓  ↓    ↓       ↓   strike price
    # // ↓  ↓    ↓       ↓   ↓
    # // ↓  ↓    ↓       ↓   ↓   type, put (P) or call (C)
    # // ↓  ↓    ↓       ↓   ↓   ↓
    # 'BTC/USDT:BTC-211225-60000-P'  // BTC/USDT put option contract strike price 60000 USDT settled in BTC (inverse)
    # on 2021-12-25
    # 'ETH/USDT:USDT-211225-40000-C' // BTC/USDT call option contract strike price 40000 USDT settled in USDT (linear,
    # vanilla) on 2021-12-25
    # 'ETH/USDT:ETH-210625-5000-P'   // ETH/USDT put option contract strike price 5000 USDT settled in ETH (inverse)
    # on 2021-06-25
    # 'ETH/USDT:USDT-210625-5000-C'  // ETH/USDT call option contract strike price 5000 USDT settled in USDT (linear,
    # vanilla) on 2021-06-25

    def __init__(
        self,
        symbol_str,
        market_separator=octobot_commons.MARKET_SEPARATOR,
        settlement_separator=octobot_commons.SETTLEMENT_ASSET_SEPARATOR,
    ):
        self.symbol_str = symbol_str
        self.base = None
        self.quote = None
        self.settlement_asset = None
        self.identifier = None
        self.strike_price = None
        self.option_type = None
        self.market_separator = market_separator
        self.settlement_separator = settlement_separator
        self.parse_symbol(self.symbol_str)

    def parse_symbol(self, symbol_str):
        """
        Parse the specified symbol
        :param symbol_str: the symbol to parse
        """
        if self.settlement_separator in symbol_str:
            (
                self.base,
                self.quote,
                self.settlement_asset,
                self.identifier,
                self.strike_price,
                self.option_type,
            ) = _parse_symbol_full(self.FULL_SYMBOL_GROUPS_REGEX, symbol_str)
        else:
            # simple (probably spot) pair, use str.split as it is much faster
            self.base, self.quote = _parse_spot_symbol(
                self.market_separator, symbol_str
            )
            self.settlement_asset = (
                self.identifier
            ) = self.strike_price = self.option_type = ""

    def base_and_quote(self):
        """
        return a tuple made of this symbol's base and quote assets
        """
        return self.base, self.quote

    def merged_str_symbol(
        self,
        market_separator=octobot_commons.MARKET_SEPARATOR,
        settlement_separator=octobot_commons.SETTLEMENT_ASSET_SEPARATOR,
    ):
        """
        return the base/quote representation of this symbol. includes settlement asset if set
        """
        if self.settlement_asset:
            return f"{self.base}{market_separator}{self.quote}{settlement_separator}{self.settlement_asset}"
        return f"{self.base}{market_separator}{self.quote}"

    def merged_str_base_and_quote_only_symbol(
        self,
        market_separator=octobot_commons.MARKET_SEPARATOR,
    ):
        """
        return the base/quote representation of this symbol. includes settlement asset if set
        """
        return f"{self.base}{market_separator}{self.quote}"

    def is_perpetual_future(self):
        """
        return True when this symbol is related to a perpetual future contract
        """
        return self.settlement_asset and not (
            self.identifier or self.strike_price or self.option_type
        )

    def is_spot(self):
        """
        return True when this symbol is related to a spot asset
        """
        return self.base and self.quote and not self.settlement_asset

    def is_future(self):
        """
        return True when this symbol is related to a non-perpetual future contract
        """
        return bool(
            self.settlement_asset and not (self.strike_price or self.option_type)
        )

    def is_option(self):
        """
        return True when this symbol is related to an option contract
        """
        return bool(
            self.settlement_asset
            and self.identifier
            and self.strike_price
            and self.option_type
        )

    def is_linear(self):
        """
        return True when this symbol is related to a linear contract based on the settlement_asset
        """
        return self.quote == self.settlement_asset if self.settlement_asset else True

    def is_inverse(self):
        """
        return True when this symbol is related to an inverse contract based on the settlement_asset
        """
        return self.base == self.settlement_asset if self.settlement_asset else False

    def is_same_base_and_quote(self, other):
        """*
        :return: True if the given symbol has the same base and quote as self
        """
        return self.base == other.base and self.quote == other.quote

    def __eq__(self, other):
        return self is other or (
            isinstance(other, Symbol)
            and self.symbol_str == other.symbol_str
            and self.base == other.base
            and self.quote == other.quote
            and self.settlement_asset == other.settlement_asset
            and self.identifier == other.identifier
            and self.strike_price == other.strike_price
            and self.option_type == other.option_type
        )

    def __str__(self):
        return self.symbol_str


@functools.lru_cache(maxsize=None)
def _parse_symbol_full(full_symbol_regex, symbol_str):
    return re.search(full_symbol_regex, symbol_str).groups()


@functools.lru_cache(maxsize=None)
def _parse_spot_symbol(separator, symbol_str):
    return symbol_str.split(separator)



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.databases.implementations.cache_database as cache_database
import octobot_commons.enums as commons_enums
import octobot_commons.errors as errors


class CacheTimestampDatabase(cache_database.CacheDatabase):
    async def get(
        self,
        timestamp: float,
        name: str = commons_enums.CacheDatabaseColumns.VALUE.value,
    ) -> dict:
        """
        Returns the value associated to the given timestamp
        :param timestamp: timestamp to get data for
        :param name: identifier of the value to get, default is commons_enums.CacheDatabaseColumns.VALUE.value
        """
        try:
            return await self._get_from_local_cache(
                commons_enums.CacheDatabaseColumns.TIMESTAMP.value, timestamp, name
            )
        except KeyError as err:
            raise errors.NoCacheValue(
                f"No cache value associated to {timestamp}"
                if err.args[0] == timestamp
                else f"No {name} value associated to {timestamp} cache."
            )

    async def get_values(
        self,
        timestamp: float,
        name: str = commons_enums.CacheDatabaseColumns.VALUE.value,
        limit=-1,
        min_timestamp=0,
    ) -> list:
        """
        Returns all the values up to the given timestamp
        :param timestamp: last timestamp to read get data to
        :param name: identifier of the value to get, default is commons_enums.CacheDatabaseColumns.VALUE.value
        :param limit: maximum number of elements to return
        :param min_timestamp: timestamp to start returning data from
        """
        try:
            await self._ensure_local_cache(
                commons_enums.CacheDatabaseColumns.TIMESTAMP.value
            )
            values = [
                values[name]
                for value_timestamp, values in self._local_cache.items()
                if min_timestamp <= value_timestamp <= timestamp and name in values
            ]
            if limit != -1:
                return values[-limit:]
            return values
        except IndexError:
            raise errors.NoCacheValue(f"No cache value associated to {name}")
        except KeyError:
            raise errors.NoCacheValue(f"No {name} value associated to {name} cache.")

    async def set(
        self,
        timestamp: float,
        value,
        name: str = commons_enums.CacheDatabaseColumns.VALUE.value,
    ) -> None:
        """
        Sets a value at the given timestamp associated to the given identifier
        :param timestamp: timestamp to set data to
        :param value: value to set
        :param name: identifier of the value to set, default is commons_enums.CacheDatabaseColumns.VALUE.value
        """
        await self._ensure_metadata()
        saved_value = self.get_serializable_value(value)
        if await self._needs_update(
            commons_enums.CacheDatabaseColumns.TIMESTAMP.value,
            timestamp,
            name,
            saved_value,
        ):
            uuid = None
            set_value = {
                commons_enums.CacheDatabaseColumns.TIMESTAMP.value: timestamp,
                name: saved_value,
            }
            if timestamp in self._local_cache:
                # set uuid in case this value already exist in db
                uuid = self._local_cache[timestamp].get(self.UUID_KEY)
                self._local_cache[timestamp][
                    commons_enums.CacheDatabaseColumns.TIMESTAMP.value
                ] = timestamp
                self._local_cache[timestamp][name] = saved_value
            else:
                self._local_cache[timestamp] = set_value
            await self.upsert(
                self.CACHE_TABLE,
                set_value,
                None,
                uuid=uuid,
                cache_query={
                    commons_enums.CacheDatabaseColumns.TIMESTAMP.value: timestamp
                },
            )

    async def set_values(
        self,
        timestamps,
        values,
        name: str = commons_enums.CacheDatabaseColumns.VALUE.value,
        additional_values_by_key: dict = None,
    ) -> None:
        """
        Sets values at the given timestamps associated to the given identifiers
        :param timestamps: timestamps to set data to
        :param values: value to set
        :param name: identifier of the values to set, default is commons_enums.CacheDatabaseColumns.VALUE.value
        :param additional_values_by_key: other key/values to set a these timestamps
        """
        await self._ensure_local_cache(
            commons_enums.CacheDatabaseColumns.TIMESTAMP.value
        )
        to_bulk_update = {
            name: [self.get_serializable_value(value) for value in values]
        }
        if additional_values_by_key:
            to_bulk_update.update(
                {
                    key: [self.get_serializable_value(value) for value in values]
                    for key, values in additional_values_by_key.items()
                }
            )
        # use optimized multiple insert to speed up the database insert operation
        await self._bulk_update_values(timestamps, to_bulk_update)

    async def _bulk_update_values(self, timestamps, to_bulk_update):
        await self._ensure_metadata()
        rows = []
        can_just_insert_data = True
        key = None
        try:
            # try to write data in the scenario their timestamp is not in cache already: can insert directly
            for index, timestamp in enumerate(timestamps):
                if timestamp in self._local_cache:
                    row = self._local_cache[timestamp]
                    # will have to update data
                    can_just_insert_data = False
                else:
                    row = {
                        commons_enums.CacheDatabaseColumns.TIMESTAMP.value: timestamp
                    }
                for key, values in to_bulk_update.items():
                    row[key] = values[index]
                self._local_cache[timestamp] = row
                rows.append(row)
            if can_just_insert_data:
                await self.log_many(self.CACHE_TABLE, rows)
            else:
                await self._update_full_database()
        except IndexError:
            raise RuntimeError(
                f"Data to set are required to have the same length as the timestamps list. "
                f"Error on the {key} values"
            )

    async def _update_full_database(self):
        # to be called to avoid multiple upsert / update which can be very slow: take full advantage of multiple inserts
        # 1. recreate all database elements from self._local_cache
        all_rows = []
        for element in self._local_cache.values():
            # remove artificial data if any
            element.pop(self.UUID_KEY, None)
            all_rows.append(element)
        # 2. delete database content
        await self.delete_all(self.CACHE_TABLE)
        # 3. insert all local cache
        await self.log_many(self.CACHE_TABLE, all_rows)
        # 4. reset self._local_cache
        await self._ensure_local_cache(
            commons_enums.CacheDatabaseColumns.TIMESTAMP.value, update=True
        )

    async def _timestamp_query(self, timestamp):
        return (await self._database.query_factory()).t == timestamp

    async def get_cache(self):
        """
        :return: the sorted read cache values
        """
        # relies on the fact that python dicts keep order
        return sorted(
            await self._database.select(self.CACHE_TABLE, None),
            key=lambda x: x[commons_enums.CacheDatabaseColumns.TIMESTAMP.value],
        )



#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.enums

from tests.signals import signal, signal_bundle


def test_to_dict(signal_bundle, signal):
    assert signal_bundle.to_dict() == {
        octobot_commons.enums.SignalBundlesAttrs.IDENTIFIER.value: "hello identifier",
        octobot_commons.enums.SignalBundlesAttrs.SIGNALS.value: [signal.to_dict()],
        octobot_commons.enums.SignalBundlesAttrs.VERSION.value: "version",
    }


def test__str__(signal_bundle):
    assert all(sub_str in str(signal_bundle)
               for sub_str in ("hello identifier", "version", "hello topic", "hi", "plop"))


def test_get_version(signal_bundle):
    assert signal_bundle._get_version() == "1.0.0"
